@@ -73,6 +73,7 @@ would appreciate credit if this program or parts of it are used.
 
 #include "tcl.h"
 #include "string.h"
+#include "expect_tcl.h"
 #include "exp_rename.h"
 #include "exp_prog.h"
 #include "exp_command.h"
@@ -147,15 +148,17 @@ RETSIGTYPE (*traps[])();
 /* Do not terminate format strings with \n!!! */
 /*VARARGS*/
 void
-exp_error(va_alist)
-va_dcl
+exp_error EXP_VARARGS_DEF(Tcl_Interp *,arg1)
+/*exp_error(va_alist)*/
+/*va_dcl*/
 {
 	Tcl_Interp *interp;
 	char *fmt;
 	va_list args;
 
-	va_start(args);
-	interp = va_arg(args,Tcl_Interp *);
+	interp = EXP_VARARGS_START(Tcl_Interp *,arg1,args);
+	/*va_start(args);*/
+	/*interp = va_arg(args,Tcl_Interp *);*/
 	fmt = va_arg(args,char *);
 	vsprintf(interp->result,fmt,args);
 	va_end(args);
@@ -751,7 +754,36 @@ when trapping, see below in child half of fork */
 		}
 
 		if (0 > (master = getptymaster())) {
-			exp_error(interp,"too many programs spawned? - out of ptys");
+			/*
+			 * failed to allocate pty, try and figure out why
+			 * so we can suggest to user what to do about it.
+			 */
+
+			int count;
+			int testfd;
+
+			if (exp_pty_error) {
+				exp_error(interp,"%s",exp_pty_error);
+				return TCL_ERROR;
+			}
+
+			count = 0;
+			for (i=3;i<exp_fd_max;i++) {
+				count += exp_fs[i].valid;
+			}
+			if (count > 10) {
+				exp_error(interp,"The system only has a finite number of ptys and you have many of them in use.  The usual reason for this is that you forgot (or didn't know) to call \"wait\" after closing each of them.");
+				return TCL_ERROR;
+			}
+
+			testfd = open("/",0);
+			close(testfd);
+
+			if (testfd != -1) {
+				exp_error(interp,"The system has no more ptys.  Ask your system administrator to create more.");
+			} else {
+				exp_error(interp,"- You have too many files are open.  Close some files or increase your per-process descriptor limit.");
+			}
 			return(TCL_ERROR);
 		}
 #ifdef PTYTRAP_DIES
@@ -865,7 +897,7 @@ when trapping, see below in child half of fork */
 
 			if (0 > (f->slave_fd = getptyslave(ttycopy,ttyinit,
 					stty_init))) {
-				exp_error("open(slave pty): %s\r\n",strerror(errno));
+				exp_error(interp,"open(slave pty): %s\r\n",strerror(errno));
 				return TCL_ERROR;
 			}
 
@@ -966,7 +998,9 @@ when trapping, see below in child half of fork */
 		 */ 
 
 		debuglog("parent: waiting for sync byte\r\n");
-		rc = read(sync_fds[0],&sync_byte,1);
+		while (((rc = read(sync_fds[0],&sync_byte,1)) < 0) && (errno == EINTR)) {
+			/* empty */;
+		}
 		if (rc == -1) {
 			errorlog("parent sync byte read: %s\r\n",strerror(errno));
 			exit(-1);
@@ -1204,7 +1238,11 @@ parent_error:
 	/* debuglog("child: waiting for go ahead from parent\r\n"); */
 
 /*	close(master);	/* force master-side close so we can read */
-	rc = read(sync2_fds[0],&sync_byte,1);
+
+	while (((rc = read(sync2_fds[0],&sync_byte,1)) < 0) && (errno == EINTR)) {
+		/* empty */;
+	}
+
 	if (rc == -1) {
 		restore_error_fd
 		errorlog("child: sync byte read: %s\r\n",strerror(errno));
@@ -1634,7 +1672,7 @@ Tcl_VarTraceProc *updateproc;	/* proc to invoke if indirect is written */
 		ckfree(i->value);
 	} else if (i->duration == EXP_PERMANENT) {
 		if (i->value) ckfree(i->value);
-		ckfree(i->variable);
+		if (i->variable) ckfree(i->variable);
 	}
 
 	i->next = exp_i_pool;
@@ -2713,6 +2751,9 @@ char **argv;
 		/* are marked sys_waited already */
 		if (!f->sys_waited) {
 			if (nowait) {
+				/* should probably generate an error */
+				/* if SIGCHLD is trapped. */
+
 				/* pass to Tcl, so it can do wait */
 				/* in background */
 				Tcl_DetachPids(1,&f->pid);
@@ -2741,6 +2782,7 @@ char **argv;
 		 */
 
 		Tcl_ReapDetachedProcs();
+		exp_rearm_sigchld(interp); /* new */
 	} else {
 		/* wait for any of our own spawned processes */
 		/* we call waitpid rather than wait to avoid running into */
@@ -2782,8 +2824,8 @@ char **argv;
 		if (m > exp_fd_max) {
 			result = NO_CHILD;	/* no children */
 			Tcl_ReapDetachedProcs();
-			exp_rearm_sigchld(interp);
 		}
+		exp_rearm_sigchld(interp);
 	}
 
 	/*  sigh, wedge forked_proc into an exp_f structure so we don't
